/*
 * $Id: $
 * $Date: $
 *
 * Author:	Kenneth Chin-Purcell, AHPCRC
 *		Copyright @ 1992, Minnesota Supercomputer Center, Inc.
 *
 * Purpose:	Magnifier
 *
 * RESTRICTED RIGHTS LEGEND
 *
 * Use, duplication, or disclosure of this software and its documentation
 * by the Government is subject to restrictions as set forth in subdivision
 * { (b) (3) (ii) } of the Rights in Technical Data and Computer Software
 * clause at 52.227-7013.
 *
 * $Log:$
 */

/* Puff is decomposed into two pieces, bound by the routines 
 * in this file.  The first part is the Puff widget, which will
 * magnify a region of the screen.  The second part is Snoop,
 * which presents a false X server and listens for graphics
 * requests by X clients.
 *
 * The location of interest for the Puff widget is set in three 
 * ways.  Motion events are selected for on the root window,
 * and these notifications are passed on the Puff widget.
 * Graphics requests in Snoop are also detected and passed
 * on to Puff.  <Ctrl>arrow-keys are also grabbed and used
 * for scanning the screen.
 *
 * In addition, the Puff widget will periodically update on its
 * own.  If the last location set was from the pointer, it will
 * continue to track the pointer.  This permits drag sequences to
 * be monitored.  Because other clients often grab the pointer
 * during drags, this is the only way for Puff to keep up.
 *
 * Unless *you* have a better way?
 */


#include "util.h"
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include "snoop.h"
#include "Puff.h"
#include <unistd.h>
#include <signal.h>


/* Change KEYMASK if grabbing <Ctrl>arrow-key is a bit too
 * obnoxious.
 */
#define KEYMASK	ControlMask


/* Global data
 */
typedef struct _PuffApp {
    XtAppContext	app;	/* the mysterious application context */
    Display		*dpy;	/* display for both Puff and Snoop */
    Window		root;	/* root window to watch */
    Widget		puff;	/* the handy Puff widget */
    Widget		shell;	/* puff's parent */
    int			dnum;	/* display number for Snoop */
    int			help;	/* a little guidence needed? */
} PuffApp;

static PuffApp		pa;

static XtResource resources[] = {
    { "displayNumber",	"DisplayNumber",	XtRInt, sizeof(int),
	  XtOffsetOf(PuffApp, dnum), XtRImmediate, (XtPointer) 1 },
    { "help",		"Help",			XtRInt, sizeof(int),
	  XtOffsetOf(PuffApp, help), XtRImmediate, (XtPointer) 0 },
};

static  XrmOptionDescRec options[] = {
    { "-dnum",	"displayNumber",XrmoptionSepArg,	NULL },
    { "-d",	"displayNumber",XrmoptionSepArg,	NULL },
    { "-help",	"help",		XrmoptionNoArg,		"1"  },
    { "-h",	"help",		XrmoptionNoArg,		"1"  },
    { "-mag",	"*Mag",		XrmoptionSepArg,	NULL },
    { "-m",	"*Mag",		XrmoptionSepArg,	NULL },
    { "-update","*delayMax",	XrmoptionSepArg,	NULL },
    { "-u",	"*delayMax",	XrmoptionSepArg,	NULL },
    { "-cwidth","*cursorWidth",	XrmoptionSepArg,	NULL },
    { "-c",	"*cursorWidth",	XrmoptionSepArg,	NULL },
};


static void Usage(void)
{
    Error("");
    Error("usage: puff [ options ]");
    Error("   	-dnum | -d  #		Display number for false server (1)");
    Error("   	-mag | -m  #		Magnification factor (6)");
    Error("   	-update | -u  #		Update delay in milliseconds (50)");
    Error("   	-cwidth | -c  #		Cursor square size in puff window (3)");
    Error("	-help | -h		This message");
    Error("");
}


/* Some tedious error handling to clean up shared memory.
 * After clearing the shared memory segments in Puff, call 
 * the previously installed error handler.
 */

static XtErrorHandler	xtHandler;
static XErrorHandler	xHandler;
static XIOErrorHandler	xIOHandler;


static void CatchXtError(String s)
{
    PuffReleaseShm(pa.puff);
    xtHandler(s);
}


static int CatchXError(Display *dpy, XErrorEvent *ee)
{
    if (PuffCatchError(pa.puff, ee))
	return 0;

    PuffReleaseShm(pa.puff);
    xHandler(dpy, ee);
    return 0;
}


static int CatchXIOError(Display *dpy)
{
    PuffReleaseShm(pa.puff);
    xIOHandler(dpy);
    return 0;
}


static void CatchSignal(int signo)
{
    PuffReleaseShm(pa.puff);
    (void) kill(getpid(), signo);
}


static void SetErrorHandlers(void)
{
    int		i;
    void	(*oldcatcher)();

    xtHandler = XtAppSetErrorHandler(pa.app, CatchXtError);
    xHandler = XSetErrorHandler(CatchXError);
    xIOHandler = XSetIOErrorHandler(CatchXIOError);

    for (i = 0; i <= SIGTERM; i++)
	switch (i) {
	  case SIGSTOP:		/* don't choke on shell signals */
	  case SIGTSTP:
	  case SIGCONT:
	    break;
	  case SIGPIPE:
	    (void) signal(i, SIG_IGN);
	    break;
	  default:
	    oldcatcher = signal(i, CatchSignal);
	    if (oldcatcher == SIG_IGN)
		(void) signal(i, SIG_IGN);	/* oops, sorry... */
	    break;
	}
}


static int ScanEvent(XEvent *event)
{
    /* Check to see if this is pointer motion on the root
     * or an arrow key grab.
     * Return TRUE to indicate this event is not for Xt.
     * Return FALSE for Xt to process event as normal.	
     */

    if (event->type == KeyPress  &&  event->xkey.window == pa.root) {

	/* An arrow key grab.  Bump the view window around.
	 */
	KeySym	ks = XKeycodeToKeysym(event->xkey.display, 
				      event->xkey.keycode, 0);
	if (ks == XK_Up)
	    PuffBump(pa.puff, PUFF_UP);
	else if (ks == XK_Down)
	    PuffBump(pa.puff, PUFF_DOWN);
	else if (ks == XK_Right)
	    PuffBump(pa.puff, PUFF_RIGHT);
	else if (ks == XK_Left)
	    PuffBump(pa.puff, PUFF_LEFT);
	else if (ks == XK_Return)
	    PuffBump(pa.puff, PUFF_RETURN);
	
	return event->xkey.window != pa.root;
    }
    
    if (event->type == MotionNotify) {
	
	/* Some Motion events have filtered down to the root window.
	 * Move the view window to the pointer.
	 */
	XEvent		nextEvent;
	
	/* Compress motion events
	 */
	while (XPending(event->xmotion.display)) {
	    XPeekEvent(event->xmotion.display, &nextEvent);
	    if (nextEvent.type == MotionNotify)
		XNextEvent(event->xmotion.display, event);
	    else
		break;
	}
	
	/* Notify Puff
	 */
	PuffSetLocation(pa.puff, pa.root, event->xmotion.x_root, 
			event->xmotion.y_root, TRUE);
	return 1;
    }
    
    return 0;
}


static void SelectInput(void)
{
    XSelectInput(pa.dpy, pa.root, PointerMotionMask | KeyPressMask);
    XGrabKey(pa.dpy, XKeysymToKeycode(pa.dpy, XK_Up),
	     KEYMASK, pa.root, TRUE, GrabModeAsync, GrabModeAsync);
    XGrabKey(pa.dpy, XKeysymToKeycode(pa.dpy, XK_Down),
	     KEYMASK, pa.root, TRUE, GrabModeAsync, GrabModeAsync);
    XGrabKey(pa.dpy, XKeysymToKeycode(pa.dpy, XK_Right),
	     KEYMASK, pa.root, TRUE, GrabModeAsync, GrabModeAsync);
    XGrabKey(pa.dpy, XKeysymToKeycode(pa.dpy, XK_Left),
	     KEYMASK, pa.root, TRUE, GrabModeAsync, GrabModeAsync);
    XGrabKey(pa.dpy, XKeysymToKeycode(pa.dpy, XK_Return),
	     KEYMASK, pa.root, TRUE, GrabModeAsync, GrabModeAsync);
}


main(int argc, char *argv[])
{
    XEvent	event;
#if XtSpecificationRelease <= 4
    Cardinal	narg = argc;
#else
    int		narg = argc;
#endif
    
    /* Create top level shell
     */
    pa.shell = XtAppInitialize(&pa.app, "Puff", options, XtNumber(options),
			       &narg, argv, NULL, NULL, 0);

    /* Get and clean up application resources
     */
    XtGetApplicationResources(pa.shell, (XtPointer) &pa, 
			      resources, XtNumber(resources), NULL, 0);
    if (pa.help) {
	Usage();
	return 0;
    }
    if (pa.dnum < 0)
	pa.dnum = 0;

    /* Create Puff widget
     */
    pa.puff  = XtCreateManagedWidget("puff", puffWidgetClass, pa.shell,
				     NULL, 0);
    pa.root  = RootWindowOfScreen(XtScreen(pa.shell));
    pa.dpy   = XtDisplay(pa.shell);
    
    CreateSnoop(pa.puff, pa.dnum);
    SetErrorHandlers();
    XtRealizeWidget(pa.shell);
    SelectInput();
    
    /* Puff will unrealize itself to signal 'quit'
     */
    while (XtIsRealized(pa.puff)) {
        XtAppNextEvent(pa.app, &event);
	if (!ScanEvent(&event))		/* Check event before passing on */
	    XtDispatchEvent(&event);
    }
    
    /* It is important to destroy the Puff widget in order
     * to free shared memory segments.
     */
    XtDestroyWidget(pa.puff);
    
    return 0;
}
