/*
 * $Id: $
 * $Date: $
 *
 * Author:	Kenneth Chin-Purcell, AHPCRC
 *		Copyright @ 1992, Minnesota Supercomputer Center, Inc.
 *
 * Purpose:	snooper 
 *
 * RESTRICTED RIGHTS LEGEND
 *
 * Use, duplication, or disclosure of this software and its documentation
 * by the Government is subject to restrictions as set forth in subdivision
 * { (b) (3) (ii) } of the Rights in Technical Data and Computer Software
 * clause at 52.227-7013.
 *
 * $Log:$
 */

/* A lot of this code has been either inspired or directly 
 * lifted from xscope.  
 *
 * Snoop listens as a 'server' at TCP port 6001 (by default), 
 * and also on a unix socket, on *this* machine, and passes 
 * on X connections to the same display that the Puff widget
 * is attached to.
 *
 * Once a connection has been accepted, data is transmitted unaltered
 * betweeen clients and the real X server.  The X requests of
 * the client are examined for graphics requests, and the first
 * x-y location in any graphics or text request is told to 
 * the Puff widget.
 *
 * X server events, responses and errors pass unexamined.
 *
 * Unlike xscope, Snoop uses Xt to handle selecting input.
 * Works like a charm.
 */
 

#include <X11/Intrinsic.h>
#include <X11/Xos.h>

#include "util.h"
#include "snoop.h"
#include "Puff.h"

#include <netdb.h>		/* struct servent and struct hostent */
#include <errno.h>		/* EINTR, EADDRINUSE, ... */
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>		/* struct iovec, used by socket.h */
#include <sys/stat.h>
#include <sys/uio.h>		/* struct iovec, used by socket.h */
#include <sys/socket.h>		/* AF_INET, SOCK_STREAM, ... */
#include <sys/un.h>		/* Unix domain: struct sockaddr_un */
#include <netinet/in.h>		/* Internet domain: struct sockaddr_in */
#include <netinet/tcp.h>	/* TCP_NODELAY */
#include <arpa/inet.h>

/* removed to compile on linux
extern int	atoi(char *);
extern int	gethostname(char *, int);
extern int	getdtablesize(void);
extern void	bzero(VoidPtr, int);
extern int	select (int, fd_set *, fd_set *, fd_set *, struct timeval *); */

#define SAVEBUMP	8000


/* The following three routines pluck longs, shorts
 * and bytes from X request records.
 */

static unsigned long ILong(char *buf, int lsbFirst)
{
    unsigned char	*c = (unsigned char *) buf;

    if (lsbFirst)
	return ((unsigned long) c[3] << 24  |
		(unsigned long) c[2] << 16  |
		(unsigned long) c[1] << 8   |
		(unsigned long) c[0]);
    return ((unsigned long) c[0] << 24  |
	    (unsigned long) c[1] << 16  |
	    (unsigned long) c[2] << 8   |
	    (unsigned long) c[3]);
}


static unsigned int IShort(char *buf, int lsbFirst)
{
    unsigned char	*c = (unsigned char *) buf;

    if (lsbFirst)
	return ((unsigned int) c[1] << 8  |
		(unsigned int) c[0]);
    return ((unsigned int) c[0] << 8  | 
	    (unsigned int) c[1]);
}


static unsigned int IByte(char *buf)
{
    return (unsigned char) buf[0];
}


/* Listen to both TCP and Unix domain sockets.
 * Set the socket so it doesn't linger after a close,
 * and is bale to reuse the port.
 */

static struct linger	dropIt = { 0, 0 };

#define X_TCP_PORT		6000

static int ListenTCP(int port)
{
    /* Listen for TCP connections on the hosts that this program
     * is being run on (INADDR_ANY).  We want to be able to reuse
     * this port again, and don't linger if the connection is dropped.
     */
    int			cs;
    struct sockaddr_in	sin;
    int			one = 1;
    
    ZeroType(&sin, sin);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = X_TCP_PORT + port;
    sin.sin_port = htons(sin.sin_port);
    
    cs = socket(sin.sin_family, SOCK_STREAM, 0);
    Verify(cs >= 0, "socket");

    (void) setsockopt(cs, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

    Verify(bind(cs, (struct sockaddr *) &sin, sizeof(sin)) == 0, "bind");
    
    (void) setsockopt(cs, SOL_SOCKET, SO_LINGER, &dropIt, sizeof(dropIt));

    Verify(listen(cs, 5) >= 0, "listen");

    return cs;
}


#define X_UNIX_DIR      "/tmp/.X11-unix"
#define X_UNIX_PATH     "/tmp/.X11-unix/X"

static int ListenUNIX(int port)
{
    /* Listen for a UNIX socket connection on the local host.
     */
    int			cs;
    struct sockaddr_un	sunix;
    int			sunixLength;
    mode_t		omask;
    
    omask = umask(0);
    
    ZeroType(&sunix, sunix);
    sunix.sun_family = AF_UNIX;
    
    (void) mkdir(X_UNIX_DIR, 0777);
    (void) chmod(X_UNIX_DIR, 0777);

    (void) sprintf(sunix.sun_path, "%s%d", X_UNIX_PATH, port);
    (void) unlink(sunix.sun_path);

    cs = socket(AF_UNIX, SOCK_STREAM, 0);
    Verify(cs >= 0, "socket");
    
    sunixLength = sizeof(sunix.sun_family) + strlen(sunix.sun_path);
    Verify(bind(cs, (struct sockaddr *)&sunix, sunixLength) == 0, "bind");

    Verify(listen(cs, 5) >= 0, "listen");

    (void) umask(omask);

    return cs;
}


static int ConnectToClient(int cs)
{
    /* A listening socket, either TCP or UNIX, has come alive.
     * Accept the connection.  For TCP sockets, turn off coalescence.
     * Set the fd to be nonblocking.
     */
    int			cfd;
    struct sockaddr	from;
    int    		len = sizeof(from);
    int			one = 1;
    
    cfd = accept(cs, &from, &len);
    Verify(cfd >= 0, "Can't connect to Client");
    
    if (from.sa_family == AF_INET)
	(void) setsockopt(cfd, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));
    (void) fcntl(cfd, F_SETFL, FNDELAY);
    
    return cfd;
}


#define INVALID_INETADDR ((unsigned long) -1)

static int ConnectToServer(char *dname)
{
    /* Go find the server, using the same display string that the Xt
     * program is using.  At this stage widgets exist, so the display
     * string and server should be valid for a new connection.  For
     * TCP sockets, turn off coalescence.  Set the fd to be nonblocking.
     */
    int		sfd;
    char	*host = NewString(dname, 0);
    int		dnum = 0;
    
    while (host[dnum] != ':'  &&  host[dnum] != '\0')
	dnum++;
    Verify(host[dnum] == ':', "Invalid display string!");
    
    host[dnum] = '\0';
    dnum = atoi(host + dnum + 1);

    if (strcmp(host, "") == 0  ||
	strcmp(host, "unix") == 0 ||
	strcmp(host, "local") == 0) {
	
	/* Make a UNIX domain connection with server (faster)
	 */
	struct sockaddr_un	sunix;

	sunix.sun_family = AF_UNIX;
	(void) sprintf (sunix.sun_path, "%s%d", X_UNIX_PATH, dnum);

	sfd = socket(AF_UNIX, SOCK_STREAM, 0);
	Verify(sfd >= 0, "socket() to Server failed");

	Verify(connect(sfd, (struct sockaddr *)&sunix, 
		       strlen(sunix.sun_path) + sizeof(sunix.sun_family)) == 0,
	       "server connect");

    } else {
	
	/* Make a TCP connection with server
	 */
	struct sockaddr_in	sin;
	int			one = 1;
	unsigned long		hostAddr;
	struct hostent		*hp; 

	ZeroType(&sin, sin);
	sin.sin_family = AF_INET;
	
	if (isascii(host[0]) && isdigit(host[0]))
	    hostAddr = inet_addr(host);
	else
	    hostAddr = INVALID_INETADDR;

	if (hostAddr == INVALID_INETADDR) {
	    hp = gethostbyname(host);
	    Verify(hp, "Host not found!");
	    Verify(hp->h_addrtype == AF_INET, "Not an Internet host!");

	    (void) memcpy(&sin.sin_addr, hp->h_addr, sizeof(sin.sin_addr));

	} else
	    sin.sin_addr.s_addr = hostAddr;
	    
	sin.sin_port = X_TCP_PORT + dnum;
	sin.sin_port = htons(sin.sin_port);
	
	sfd = socket(sin.sin_family, SOCK_STREAM, 0);
	Verify(sfd >= 0, "socket() to Server failed");
	
	(void) setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));

	Verify(connect(sfd, (struct sockaddr *)&sin, sizeof(sin)) >= 0,
	       "Can't connect to Server");
    }

    (void) fcntl(sfd, F_SETFL, FNDELAY);
    
    return sfd;
}


static void CloseConnection(Snoop *snoop, int cfd)
{
    /* Close and clean up a client <-> server connection.
     */
    int		pfd = snoop->file[cfd].pair;

    snoop->file[cfd].Process = NULL;	/* Stops processing */
    snoop->file[pfd].Process = NULL;

    Free(snoop->file[cfd].saved);
    Free(snoop->file[pfd].saved);
    snoop->file[cfd].saved = NULL;
    snoop->file[pfd].saved = NULL;
    snoop->file[cfd].saveSize = 0;
    snoop->file[pfd].saveSize = 0;

    XtRemoveInput(snoop->file[cfd].xid);
    XtRemoveInput(snoop->file[pfd].xid);

    (void) close(cfd);
    (void) close(pfd);
}


static void DecodeRequest(Snoop *snoop, int fd, char *data)
{
    /* Given an X requst in data, search for graphics calls.
     */
    FileInfo	*fi = snoop->file + fd;
    int		request = IByte(data);
    
    switch (request) {
      case 64:
      case 65:
      case 66:
      case 67:
      case 68:
      case 69:
      case 70:
      case 71:
      case 74:
      case 75:
      case 76:
      case 77:
	/* Set Puff's location to the first x-y location
	 * in the graphics request.
	 */
	PuffSetLocation(snoop->puff, ILong(data + 4, fi->lsbFirst),
			IShort(data + 12, fi->lsbFirst),
			IShort(data + 14, fi->lsbFirst), FALSE);
	break;
      default:
	break;
    }
}


static long FinishRequest(void *closure, int fd, char *data);

static long StartRequest(void *closure, int fd, char *data)
{
    /* Find out how much more data is needed.
     */
    Snoop	*snoop = (Snoop *) closure;
    FileInfo	*fi = snoop->file + fd;
    int		used = 0;
    
    fi->Process = FinishRequest;
    fi->needed = 4 * IShort(data + 2, fi->lsbFirst);
    snoop->halfBaked = True;

    return used;
}


/*ARGSUSED*/
static long FinishRequest(void *closure, int fd, char *data)
{
    /* Decode this complete X request, and start looking 
     * for the next.
     */
    Snoop	*snoop = (Snoop *) closure;
    FileInfo	*fi = snoop->file + fd;
    int		used = fi->needed;

    fi->Process = StartRequest;
    fi->needed = 4;

    DecodeRequest(snoop, fd, data);
    snoop->halfBaked = False;

    return used;
}


static long FinishSetupMessage(void *closure, int fd, char *data);

static long StartSetupMessage(void *closure, int fd, char *data)
{
    /* Client sends 12 initial bytes and a possible
     * name, which we don;t care about.
     * The server will match the client's byte swapping.
     */
    Snoop	*snoop = (Snoop *) closure;
    FileInfo	*fi = snoop->file + fd;
    int		used = 0;
    
    fi->lsbFirst = data[0] == 'l';
    snoop->file[fi->pair].lsbFirst = fi->lsbFirst;

    fi->Process = FinishSetupMessage;
    fi->needed += (IShort(data + 6, fi->lsbFirst) * 4 + 
		   IShort(data + 8, fi->lsbFirst) * 4);
    snoop->halfBaked = True;

    return used;
}


/*ARGSUSED*/
static long FinishSetupMessage(void *closure, int fd, char *data)
{
    /* The client has finished introducing itself, so now we
     * can get down to business.
     */
    Snoop	*snoop = (Snoop *) closure;
    FileInfo	*fi = snoop->file + fd;
    int		used = fi->needed;
    
    fi->Process = StartRequest;
    fi->needed = 4;
    snoop->halfBaked = False;

    return used;
}


static void AppendData(FileInfo *fi, char *buf, long n)
{
    /* Append n bytes in buff to the fd's data buffer.
     */
    if (n > 0) {
	if (fi->nsaved + n > fi->saveSize) {
	    /* printf("%d\n", n); */
	    while (fi->nsaved + n > fi->saveSize)
		fi->saveSize += SAVEBUMP;
	    fi->saved = (char *) realloc(fi->saved, fi->saveSize);
	    MemCheck(fi->saved);
	}
	(void) memcpy(fi->saved + fi->nsaved, buf, n);
	fi->nsaved += n;
    }
}


static void PopData(FileInfo *fi, long n)
{
    /* Pop n off of the fd's data buffer.
     */
    fi->nsaved -= n;
    if (fi->nsaved < 0)
	fi->nsaved = 0;
    if (fi->nsaved > 0)
	(void) memcpy(fi->saved, fi->saved + n, fi->nsaved);
}


static void ProcessStream(Snoop *snoop, int fd, char *data, long ndata)
{
    /* Accumulate data from the client until we have a enough
     * to take action.
     */
    FileInfo	*fi = snoop->file + fd;
    char	*packet;
    long	m;

    while (fi->nsaved + ndata >= fi->needed) {

	/* There is enough data between the saved and input buffer
	 * to process a packet.  Use the saved buffer if there is data
	 * there, tacking on extra bytes if needed.  After the saved
	 * buffer is exhausted use the input buffer.
	 */

	if (fi->nsaved == 0)
	    packet = data;
	else {
	    if (fi->nsaved < fi->needed) {
		m = fi->needed - fi->nsaved;
		AppendData(fi, data, m);
		data += m;
		ndata -= m;
	    }
	    packet = fi->saved;
	}

	m = fi->Process(snoop, fd, packet);
	
	if (m > 0) {
	    if (fi->nsaved > 0)
		PopData(fi, m);
	    else {
		data += m;
		ndata -= m;
	    }
	}
    }
    
    /* Save straggler bytes for next time
     */
    if (ndata > 0)
	AppendData(fi, data, ndata);
}


static void WaitUntilWriteable(int fd)
{
    fd_set	wfds;
    
    FD_ZERO(&wfds);
    FD_SET(fd, &wfds);

    (void) select(fd + 1, (fd_set *) NULL, &wfds, (fd_set *) NULL, 
		  (struct timeval *) NULL);
}


#define BUFSIZE		2048

/*ARGSUSED*/
static void DispatchInput(XtPointer closure, int *fd, XtInputId *xid)
{
    /* Read data from the input fd, then keep writing to output
     * fd until the input buffer is dispatched.  If this is a request
     * from a client, process the stream.
     */
    Snoop		*snoop = (Snoop *) closure;
    int			ifd = *fd;
    int			ofd = snoop->file[ifd].pair;
    long    		ninput, noutput;
    long    		pending;
    char		*data;
    
    static char 	*dataBuf = NULL;

    if (!dataBuf) {
	dataBuf = CallocType(char, BUFSIZE);
	MemCheck(dataBuf);
    }
    
    do {
	data = dataBuf;
	ninput = read(ifd, data, BUFSIZE);
	if (ninput <= 0) {
	    CloseConnection(snoop, ifd);
	    return;
	}
	
	for (pending = ninput; pending > 0; ) {
	    
	    noutput = write(ofd, data, pending);
	    
	    if (noutput >= 0) {
		data += noutput;
		pending -= noutput;
    
	    } else if (errno == EWOULDBLOCK)
		WaitUntilWriteable(ofd);
	    
	    else {
		if (snoop->file[ofd].server)
		    perror("Error on write to Server");
		else
		    perror("Error on write to Client");
		CloseConnection(snoop, ifd);
		snoop->halfBaked = False;
		return;
	    }
	}

	if (snoop->file[ifd].Process)
	    ProcessStream(snoop, ifd, dataBuf, ninput);
	    
    } while (snoop->halfBaked);
}


/*ARGSUSED*/
static void NewConnection(XtPointer closure, int *fd, XtInputId *xid)
{
    /* Input is ready on one of the listening "Server" sockets.
     * Create a pair of fd's for the new connection.
     */
    Snoop	*snoop = (Snoop *) closure;
    int		cfd = ConnectToClient(*fd);
    int		sfd = ConnectToServer(DisplayString(XtDisplay(snoop->puff)));
    
    /* client */
    snoop->file[cfd].server = 0;
    snoop->file[cfd].pair = sfd;
    snoop->file[cfd].saveSize = SAVEBUMP;
    snoop->file[cfd].saved = (char *) malloc(snoop->file[cfd].saveSize);
    snoop->file[cfd].nsaved = 0;
    snoop->file[cfd].Process = StartSetupMessage;
    snoop->file[cfd].needed = 12;
    
    snoop->file[cfd].xid = XtAppAddInput(snoop->app, cfd, 
					 (XtPointer) XtInputReadMask, 
					 DispatchInput, (XtPointer) snoop);

    /* server */
    snoop->file[sfd].server = 1;
    snoop->file[sfd].pair = cfd;
    snoop->file[sfd].saveSize = SAVEBUMP;
    snoop->file[sfd].saved = (char *) malloc(snoop->file[sfd].saveSize);
    snoop->file[sfd].nsaved = 0;
    snoop->file[sfd].Process = NULL;	/* Not interested */
    snoop->file[sfd].needed = 8;

    snoop->file[sfd].xid = XtAppAddInput(snoop->app, sfd, 
					 (XtPointer) XtInputReadMask, 
					 DispatchInput, (XtPointer) snoop);
}


void CreateSnoop(Widget puff, int dnum)
{
    /* Create two listening sockets for X connections, 
     * and allocate a table to handle all file descriptors.
     */
    Snoop	*snoop = MallocType(Snoop);

    MemCheck(snoop);
    ZeroType(snoop, Snoop);

    snoop->puff = puff;
    snoop->app = XtWidgetToApplicationContext(puff);

    snoop->nfile = getdtablesize();
    snoop->file = CallocType(FileInfo, snoop->nfile);	
    MemCheck(snoop->file);
    
    snoop->halfBaked = False;

    (void) XtAppAddInput(snoop->app, ListenUNIX(dnum), 
			 (XtPointer) XtInputReadMask, 
			 NewConnection, (XtPointer) snoop);

    (void) XtAppAddInput(snoop->app, ListenTCP(dnum), 
			 (XtPointer) XtInputReadMask, 
			 NewConnection, (XtPointer) snoop);
}


