/* $Id: util.c,v 1.3 92/02/04 14:36:45 ken Exp Locker: ken $
 * $Date: 92/02/04 14:36:45 $
 *
 * Author:	Kenneth Chin-Purcell, AHPCRC
 *		Copyright @ 1992, Minnesota Supercomputer Center, Inc.
 *
 * Utilities that use basic C types and UNIX functions.
 *
 * RESTRICTED RIGHTS LEGEND
 *
 * Use, duplication, or disclosure of this software and its documentation
 * by the Government is subject to restrictions as set forth in subdivision
 * { (b) (3) (ii) } of the Rights in Technical Data and Computer Software
 * clause at 52.227-7013.
 *
 * $Log:	util.c,v $
 * Revision 1.3  92/02/04  14:36:45  ken
 * Remove wcl, cleanup
 * 
 */

#include "util.h"

#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>

extern void	abort(void);
extern void	exit(int); 
extern long	atol(char *); 


/* Print a string to standard eror.
 */
void Error(char *desc)
{
    (void) fprintf(stderr, "%s\n", desc);
}


/* Terminate program after printing an error message.
 * Use via the macros Verify and MemCheck.
 */
void BailOut(char *desc, char *fname, int lineno)
{
    char	s[256];
    
    (void) sprintf(s, "Error - %s, at %s:%d", desc, fname, lineno);
    if (errno)
	perror(s);
    else
	Error(s);
    exit(1);
}


/* Find the size of a file in bytes,
 */
long FindFileSize(char *fname)
{
    struct stat	sbuf;

    if (stat(fname, &sbuf) != 0) {
	Error("Can't stat file");
	return 0;
    }
    return sbuf.st_size;
}


/* Depending on the file length, read a 768 bytes
 * or 1024 ascii integers to create a color map.
 * Bytes ordered	rgbrgbrgb...
 * Integers ordered	index red grn blu index red grn blu ...
 */
void ReadCmap(short cmap[256][3], char *cmapName)
{
    int		ix, r, g, b;
    FILE	*cmapFile;

    cmapFile = fopen(cmapName, "r");
    Verify(cmapFile, "can't open color map file");
    
    if (FindFileSize(cmapName) == 768) {
	for (ix = 0; ix < 256; ix++) {
	    cmap[ix][0] = fgetc(cmapFile);
	    cmap[ix][1] = fgetc(cmapFile);
	    cmap[ix][2] = fgetc(cmapFile);
	}
    } else {
	while (fscanf(cmapFile, "%d%d%d%d", &ix, &r, &g, &b) == 4) {
	    if (ix < 0  ||  ix >= 256) {
		Error("Color map index out of range");
		continue;
	    }
	    cmap[ix][0] = r & 0xff;
	    cmap[ix][1] = g & 0xff;
	    cmap[ix][2] = b & 0xff;
	}
    }
    (void) fclose(cmapFile);
}


/* Create a new string from an old one, tacking on extra bytes
 */
char *NewString(char *old, int extra)
{
    int		slen;
    char	*s;

    if (old)
	slen = strlen(old) + 1 + extra;
    else
	slen = MIN(extra, 1);
    
    MemCheck(s = CallocType(char, slen));
    
    if (old)
	(void) strcpy(s, old);
    else
	s[0] = '\0';
    
    return s;
}


/* Match two tokens, ignoring white space and case.
 */
int NameMatch(char *str1, char *str2)
{
    if (!str1 || !str2)
	return 0;
    while (isspace(*str1))
	str1++;
    while (isspace(*str2))
	str2++;
    while (isalnum(*str1) && isalnum(*str2))
	if (tolower(*str1++) != tolower(*str2++))
	    return 0;
    return 1;
}


/* Return a comma separated string representation of a long.
 */
void CommaLong(char *s, long i)
{
    long	ai = ABS(i);

    if (ai < 10000)
	(void) sprintf(s, "%ld", i);
    else if (ai < 1000000)
	(void) sprintf(s, "%ld,%03ld", i / 1000, ai % 1000);
    else
	(void) sprintf(s, "%ld,%03ld,%03ld", 
		       i / 1000000, (ai % 1000000) / 1000, ai % 1000);
}


/* Find the largest even divisor of an integer.
 */
long Factor(long a)
{
    long	i;

    for (i = sqrt((double)a); i; i--)
	if (a % i == 0)
	    break;
    return i;
}


/* Round integer up to next modulo.
 */
long RoundUp(long a, long modulo)
{
    long	b = a % modulo;

    if (b)
	return a + modulo - b;
    else
	return a;
}

/* Handy array creation.
 */
long *MakeLongArray(long num)
{
    long	*larr = CallocType(long, num);
    MemCheck(larr);
    return larr;
}


float *MakeFloatArray(long num)
{
    float	*farr = CallocType(float, num);
    MemCheck(farr);
    return farr;
}


/* Routines for reading ascii files.
 * Assume comment lines begin with '#'.
 */
char *ReadString(FILE *file)
{
    static char	buf[1024];
    int		nread;

    do {
	nread = fscanf(file, "%s", buf);
	if (nread < 1)
	    return NULL;
	if (buf[0] == '#') {
	    (void) fgets(buf, 1024, file);
	    nread = 0;
	}
    } while (nread < 1);

    return buf;
}


long ReadLong(FILE *file)
{
    char	*s = ReadString(file);

    if (s)
	return atol(s);
    return 0;
}


float ReadFloat(FILE *file)
{
    char	*s = ReadString(file);

    if (s)
	return atof(s);
    return 0.0;
}


char *ReadToken(FILE *file)
{
    char	*s = ReadString(file);

    if (s)
	return s;
    return "";
}


int PositionAtString(char *s, FILE *file)
{
    char	*token;

    rewind(file);
    do {
	token = ReadString(file);
	if (!token)			/* No more strings in file */
	    return EOF;
    } while (!MATCH(token, s));
    return 0;
}


void ReadFloatArray(float *farr, long num, FILE *file)
{
    int		i;
    for (i = 0; i < num; i++)
	farr[i] = ReadFloat(file);
}


void ReadLongArray(long *larr, long num, FILE *file)
{
    int		i;
    for (i = 0; i < num; i++)
	larr[i] = ReadLong(file);
}


/* Test for endian
 */
int BigEndFirst(void)
{
    unsigned long	l = 0x01020304;
    unsigned char	*c = (unsigned char *) &l;
    
    if (*c == 4)
	return 0;
    else
	return 1;
}


int Round(float x)
{
    return floor(x + 0.4999);
}
